/*	SCCS Id: @(#)tiletext.c	3.3	1999/10/24	*/
/* NetHack may be freely redistributed.  See license for details. */

#include "config.h"
#include "tile.h"

/*
 * TEXTCOLORMAPSPACE
 *
 * This is the maximum number of possible unique colours in a .TXT file
 * file. MAXCOLORMAPSIZE may be set (in tile.h) to be smaller than this
 * in which case only that many unique colours may be present in a .TXT
 * file (and in all merged .TXT files). MAXCOLORMAPSIZE may be larger
 * than this without penalty. A value of n*TEXTCOLORMAPSPACE where n is
 * the number of .TXT files that are going to be merged (currently three;
 * monsters, objects & others) produces maximum generality. Increasing
 * MAXCOLORMAPSIZE has performance issues for the game (TEXTCOLORMAPSPACE
 * does not). Windowing systems are required to cope with MAXCOLORMAPSIZE
 * colours and map them onto their possible palette (see tile.doc) so this
 * also introduces an additional burden.
 */

#define TEXTCOLORMAPSPACE	4096

pixval ColorMap[3][MAXCOLORMAPSIZE];
int colorsinmap;
pixval MainColorMap[3][MAXCOLORMAPSIZE];
int colorsinmainmap;

static short color_index[TEXTCOLORMAPSPACE];
static int num_colors;
static char charcolors[MAXCOLORMAPSIZE][3];

static int placeholder_init = 0;
static pixel placeholder[TILE_Y][TILE_X];
static FILE *tile_file;
static int tile_set, tile_set_indx;
static char *text_sets[] = {
#if (TILE_X==8)
  "monthin.txt", "objthin.txt", "oththin.txt", "comthin.txt"
#elif (TILE_X==32 && TILE_Y==32)
  "mon32.txt", "obj32.txt", "oth32.txt", "com32.txt"
#else
  "monsters.txt", "objects.txt", "other.txt", "combined.txt"
#endif
};

extern const char *FDECL(tilename, (int, int));
static void FDECL(read_text_colormap, (FILE *));
static boolean FDECL(write_text_colormap, (FILE *));
static boolean FDECL(read_txttile, (FILE *, pixel(*)[TILE_X]));
static void FDECL(write_txttile, (FILE *, pixel(*)[TILE_X]));

/*
 * ALI
 *
 * Support for 4096 colours.
 *
 * Issues:
 *	- NetHack 3.2/3.3 .TXT readers have a bug which causes them
 *	  to overflow the "c" array in read_text_colormap if more than
 *	  one character is encountered in the set A-Za-z0-9 where it
 *	  is expecting a colour key. This is because no maximum field
 *	  width has been specified for the %[...] scan format.
 *
 * Design goals:
 *	- Capable of reading and writing 62/4096 colour files
 *	- Writes 62 files which NetHack 3.2/3.3 can read.
 *	- Writes 4096 files which NetHack 3.2/3.3 fails gracefully on.
 *
 * We achieve this by introducing two new legal characters "_" and "$".
 * This brings the number of legal characters to 64 which is a nice round
 * number and also means that as long as we arrange for the first colour
 * key in a 4096 file to start with "_", 3.2/3.3 readers will fail without
 * crashing. Instead they will see no valid colour map. The error message
 * thus generated (no colormap set yet) is admittedly not very informative.
 */

/* Ugh.  DICE doesn't like %[A-Z], so we have to spell it out... */
#define FORMAT_STRING \
"%2[_ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789$] = (%d, %d, %d) "

static char bysx2char(i)
int i;
{
	char c;
	if (!i) c = '_';
	else if (i < 27) c = 'A' + i - 1;
	else if (i < 53) c = 'a' + i - 27;
	else if (i < 63) c = '0' + i - 53;
	else c = '$';
	return c;
}

static int char2bysx(c)
char c;
{
	int i;
	if ( c == '_' ) i = 0;
	else if (c >= 'A' && c <= 'Z') i = c - 'A' + 1;
	else if (c >= 'a' && c <= 'z') i = c - 'a' + 27;
	else if (c >= '0' && c <= '9') i = c - '0' + 53;
	else if ( c == '$' ) i = 63;
	else i = -1;
	return i;
}

static void
read_text_colormap(txtfile)
FILE *txtfile;
{
	int i, n, r, g, b;
	char c[3];

	for (i = 0; i < TEXTCOLORMAPSPACE; i++)
		color_index[i] = -1;

	num_colors = 0;
	while (fscanf(txtfile, FORMAT_STRING, c, &r, &g, &b) == 4) {
		if (c[1])
		    n=char2bysx(c[0]) * 64 + char2bysx(c[1]);
		else
		    n=char2bysx(c[0]);
		if (n >= 0 && n < TEXTCOLORMAPSPACE)
			color_index[n] = num_colors;
		else
		{
			Fprintf(stderr, "error: Illegal color in colormap %s\n",
			  c);
			continue;
		}
		ColorMap[CM_RED][num_colors] = r;
		ColorMap[CM_GREEN][num_colors] = g;
		ColorMap[CM_BLUE][num_colors] = b;
		num_colors++;
	}
	colorsinmap = num_colors;
}

#undef FORMAT_STRING

static boolean
write_text_colormap(txtfile)
FILE *txtfile;
{
	int i;
	char c[3]="?";

	num_colors = colorsinmainmap;
	if (num_colors > MAXCOLORMAPSIZE) {
		Fprintf(stderr, "too many colors (%d)\n", num_colors);
		return FALSE;
	}
	for (i = 0; i < num_colors; i++) {
		if (num_colors > 62)
		{
			c[0] = bysx2char(i / 64);
			c[1] = bysx2char(i % 64);
		}
		else
			c[0] = bysx2char(i + 1);

		strcpy(charcolors[i], c);
		Fprintf(txtfile, "%s = (%d, %d, %d)\n", c,
						(int)MainColorMap[CM_RED][i],
						(int)MainColorMap[CM_GREEN][i],
						(int)MainColorMap[CM_BLUE][i]);
	}
	return TRUE;
}

static boolean
read_txttile(txtfile, pixels)
FILE *txtfile;
pixel (*pixels)[TILE_X];
{
	int ph, i, j, k, n;
	int tile_no;
	char buf[BUFSZ], ttype[BUFSZ];
	const char *p, *fmt_string;
	char c[3];

	if (fscanf(txtfile, "# %s %d (%[^)])", ttype, &i, buf) <= 0)
		return FALSE;
	
	ph = strcmp(ttype, "placeholder") == 0;

	if (!ph && strcmp(ttype,"tile") != 0)
		Fprintf(stderr,
			"Keyword \"%s\" unexpected for entry %d\n",
			ttype, i);

	if (tile_set != 0) {
	    /* check tile name, but not relative number, which will
	     * change when tiles are added
	     */
	    p = tilename(tile_set, tile_set_indx);
	    if (p && strcmp(p, buf)) {
		Fprintf(stderr, "warning: for tile %d (numbered %d) of %s,\n",
				tile_set_indx, i, text_sets[tile_set-1]);
		Fprintf(stderr, "\tfound '%s' while expecting '%s'\n",
				buf, p);
	    }
	}
	tile_no=i;
	tile_set_indx++;

	/* look for non-whitespace at each stage */
	if (fscanf(txtfile, "%1s", c) < 0) {
		Fprintf(stderr, "unexpected EOF\n");
		return FALSE;
	}
	if (c[0] != '{') {
		Fprintf(stderr, "didn't find expected '{'\n");
		return FALSE;
	}
	fmt_string = num_colors > 64 ? "%2s" : "%1s";
	for (j = 0; j < TILE_Y; j++) {
		for (i = 0; i < TILE_X; i++) {
			if (fscanf(txtfile, fmt_string, c) < 0) {
				Fprintf(stderr, "unexpected EOF\n");
				return FALSE;
			}
			if (c[1])
				n=char2bysx(c[0]) * 64 + char2bysx(c[1]);
			else
				n=char2bysx(c[0]);
			if (n >= 0 && n < TEXTCOLORMAPSPACE)
				k = color_index[n];
			else
				k = -1;
			if (k == -1)
				Fprintf(stderr,
					"color %s not in colormap!\n", c);
			else {
				pixels[j][i].r = ColorMap[CM_RED][k];
				pixels[j][i].g = ColorMap[CM_GREEN][k];
				pixels[j][i].b = ColorMap[CM_BLUE][k];
			}
		}
	}
	if ( ph ) {
	    /* remember it for later */
	    memcpy( placeholder, pixels, sizeof(placeholder) );
	}
	if (fscanf(txtfile, "%1s ", c) < 0) {
		Fprintf(stderr, "unexpected EOF\n");
		return FALSE;
	}
	if (c[0] != '}') {
		Fprintf(stderr, "didn't find expected '}'\n");
		return FALSE;
	}
#ifdef _DCC
	/* DICE again... it doesn't seem to eat whitespace after the } like
	 * it should, so we have to do so manually.
	 */
	while ((*c = fgetc(txtfile)) != EOF && isspace(*c))
		;
	ungetc(*c, txtfile);
#endif
	return TRUE;
}

static void
write_txttile(txtfile, pixels)
FILE *txtfile;
pixel (*pixels)[TILE_X];
{
	const char *p;
	const char *type;
	int i, j, k;

	if ( memcmp(placeholder, pixels, sizeof(placeholder)) == 0 )
	    type = "placeholder";
	else
	    type = "tile";

	if (tile_set == 0)
		Fprintf(txtfile, "# %s %d (unknown)\n", type, tile_set_indx);
	else {
		p = tilename(tile_set, tile_set_indx);
		if (p)
		    Fprintf(txtfile, "# %s %d (%s)\n", type, tile_set_indx, p);
		else
		    Fprintf(txtfile, "# %s %d (null)\n", type, tile_set_indx);
	}
	tile_set_indx++;

	Fprintf(txtfile, "{\n");
	for (j = 0; j < TILE_Y; j++) {
		Fprintf(txtfile, "  ");
		for (i = 0; i < TILE_X; i++) {
			for (k = 0; k < num_colors; k++) {
				if (ColorMap[CM_RED][k] == pixels[j][i].r &&
				    ColorMap[CM_GREEN][k] == pixels[j][i].g &&
				    ColorMap[CM_BLUE][k] == pixels[j][i].b)
					break;
			}
			if (k >= num_colors)
				Fprintf(stderr, "color not in colormap!\n");
			(void) fputs(charcolors[k], txtfile);
		}
		Fprintf(txtfile, "\n");
	}
	Fprintf(txtfile, "}\n");
}

/* initialize main colormap from globally accessed ColorMap */
void
init_colormap()
{
	int i;

	colorsinmainmap = colorsinmap;
	for (i = 0; i < colorsinmap; i++) {
		MainColorMap[CM_RED][i] = ColorMap[CM_RED][i];
		MainColorMap[CM_GREEN][i] = ColorMap[CM_GREEN][i];
		MainColorMap[CM_BLUE][i] = ColorMap[CM_BLUE][i];
	}
}

/* merge new colors from ColorMap into MainColorMap */
void
merge_colormap()
{
	int i, j;

	for (i = 0; i < colorsinmap; i++) {
		for (j = 0; j < colorsinmainmap; j++) {
		    if (MainColorMap[CM_RED][j] == ColorMap[CM_RED][i] &&
			MainColorMap[CM_GREEN][j] == ColorMap[CM_GREEN][i] &&
			MainColorMap[CM_BLUE][j] == ColorMap[CM_BLUE][i])
			    break;
		}
		if (j >= colorsinmainmap) {	/* new color */
		    if (colorsinmainmap >= MAXCOLORMAPSIZE) {
			Fprintf(stderr,
			    "Too many colors to merge -- excess ignored.\n");
		    }
		    j = colorsinmainmap;
		    MainColorMap[CM_RED][j] = ColorMap[CM_RED][i];
		    MainColorMap[CM_GREEN][j] = ColorMap[CM_GREEN][i];
		    MainColorMap[CM_BLUE][j] = ColorMap[CM_BLUE][i];
		    colorsinmainmap++;
		}
	}
}

boolean
fopen_text_file(filename, type)
const char *filename;
const char *type;
{
	const char *p;
	int i;

	if (tile_file != (FILE *)0) {
		Fprintf(stderr, "can only open one text file at at time\n");
		return FALSE;
	}

	tile_file = fopen(filename, type);
	if (tile_file == (FILE *)0) {
		Fprintf(stderr, "cannot open text file %s\n", filename);
		return FALSE;
	}

	p = rindex(filename, '/');
	if (p) p++;
	else p = filename;

	tile_set = 0;
	for (i = 0; i < SIZE(text_sets); i++) {
		if (!strcmp(p, text_sets[i]))
			tile_set = i+1;
	}
	tile_set_indx = 0;

	if (!strcmp(type, RDTMODE)) {
		/* Fill placeholder with noise */
		if ( !placeholder_init ) {
		    placeholder_init++;
		    for ( i=0; i<sizeof(placeholder); i++ )
			((char*)placeholder)[i]=i%256;
		}

		read_text_colormap(tile_file);
		if (!colorsinmainmap)
			init_colormap();
		else
			merge_colormap();
		return TRUE;
	} else if (!strcmp(type, WRTMODE)) {
		if (!colorsinmainmap) {
			Fprintf(stderr, "no colormap set yet\n");
			return FALSE;
		}
		return(write_text_colormap(tile_file));
	} else {
		Fprintf(stderr, "bad mode (%s) for fopen_text_file\n", type);
		return FALSE;
	}
}

boolean
read_text_tile(pixels)
pixel (*pixels)[TILE_X];
{
	return(read_txttile(tile_file, pixels));
}

boolean
write_text_tile(pixels)
pixel (*pixels)[TILE_X];
{
	write_txttile(tile_file, pixels);
	return TRUE;
}

int
fclose_text_file()
{
	int ret;

	ret = fclose(tile_file);
	tile_file = (FILE *)0;
	return ret;
}
