/* NetHack may be freely redistributed.  See license for details. */

#include <X11/xpm.h>
#include "config.h"
#include "tile.h"

#ifndef MONITOR_HEAP
extern long *FDECL(alloc, (unsigned int));
#endif

static int tiles_across, tiles_down, curr_tiles_across, curr_tiles_down;
static XpmImage image;
static XpmInfo info;

/* Maybe these should be in xpm.h, but there isn't one. */
boolean FDECL(fopen_xpm_file, (char *, char *));
boolean FDECL(read_xpm_tile, (pixel(*)[]));
int NDECL(fclose_xpm_file);

boolean
fopen_xpm_file(filename, type)
char *filename;
char *type;
{
    int i,n;
    int errorcode;
    char fmt[20];
    unsigned long r, g, b;
    const pixel bg = DEFAULT_BACKGROUND;

    if (strcmp(type, RDBMODE)) {
	Fprintf(stderr, "using reading routine for non-reading?\n");
	return FALSE;
    }
    errorcode = XpmReadFileToXpmImage(filename, &image, &info);
    if (errorcode != XpmSuccess) {
	Fprintf(stderr, "cannot open xpm file %s: %s\n", filename,
	  XpmGetErrorString(errorcode));
	return FALSE;
    }

    if (image.width % TILE_X) {
	Fprintf(stderr, "error: width %d not divisible by %d\n",
	  image.width, TILE_X);
	exit(EXIT_FAILURE);
    }
    tiles_across = image.width / TILE_X;
    curr_tiles_across = 0;
    if (image.height % TILE_Y) {
	Fprintf(stderr, "error: height %d not divisible by %d\n",
	  image.height, TILE_Y);
	/* exit(EXIT_FAILURE) */;
    }
    tiles_down = image.height / TILE_Y;
    curr_tiles_down = 0;

    for (i = 0; i < image.ncolors; ++i) {
	if (!strcmp(image.colorTable[i].c_color, "None"))
	{
	    ColorMap[CM_RED][i] = bg.r;
	    ColorMap[CM_GREEN][i] = bg.g;
	    ColorMap[CM_BLUE][i] = bg.b;
	}
	else if (image.colorTable[i].c_color[0] == '#')
	{
	    n = strlen(image.colorTable[i].c_color + 1);
	    if (n % 3 || n > 12 || n < 3)
	    {
		Fprintf(stderr, "error: Unknown color defn for %s (%s)\n",
		  image.colorTable[i].string, image.colorTable[i].c_color);
		exit(EXIT_FAILURE);
	    }
	    n /= 3;
	    sprintf(fmt, "%%0%dlx%%0%dlx%%0%dlx", n, n, n);
	    if (sscanf(image.colorTable[i].c_color + 1, fmt, &r, &g, &b) != 3)
	    {
		Fprintf(stderr, "error: Unknown color defn for %s (%s)\n",
		  image.colorTable[i].string, image.colorTable[i].c_color);
		exit(EXIT_FAILURE);
	    }
	    if (n>=2)
	    {
		ColorMap[CM_RED][i]=r>>(n*4-8);
		ColorMap[CM_GREEN][i]=g>>(n*4-8);
		ColorMap[CM_BLUE][i]=b>>(n*4-8);
	    }
	    else
	    {
		ColorMap[CM_RED][i]=r<<4;
		ColorMap[CM_GREEN][i]=g<<4;
		ColorMap[CM_BLUE][i]=b<<4;
	    }
	}
	else
	{
	    Fprintf(stderr, "error: Unknown color defn for %s (%s)\n",
	      image.colorTable[i].string, image.colorTable[i].c_color);
	    exit(EXIT_FAILURE);
	}
    }
    colorsinmap = image.ncolors;

    return TRUE;
}

/* Read a tile.  Returns FALSE when there are no more tiles */
boolean
read_xpm_tile(pixels)
pixel (*pixels)[TILE_X];
{
    int i, j;
    unsigned int *src;

    if (curr_tiles_down >= tiles_down) return FALSE;
    if (curr_tiles_across == tiles_across) {
	curr_tiles_across = 0;
	curr_tiles_down++;
	if (curr_tiles_down >= tiles_down) return FALSE;
    }
    src = image.data + curr_tiles_down * TILE_Y * image.width +
      curr_tiles_across * TILE_X;
    for (j = 0; j < TILE_Y; j++) {
	for (i = 0; i < TILE_X; i++) {
	    pixels[j][i].r = ColorMap[CM_RED][src[j * image.width + i]];
	    pixels[j][i].g = ColorMap[CM_GREEN][src[j * image.width + i]];
	    pixels[j][i].b = ColorMap[CM_BLUE][src[j * image.width + i]];
	}
    }
    curr_tiles_across++;

    /* check for "filler" tile */
    for (j = 0; j < TILE_Y; j++) {
	for (i = 0; i < TILE_X && i < 4; i += 2) {
	    if (pixels[j][i].r != ColorMap[CM_RED][0] ||
		pixels[j][i].g != ColorMap[CM_GREEN][0] ||
		pixels[j][i].b != ColorMap[CM_BLUE][0] ||
		pixels[j][i+1].r != ColorMap[CM_RED][1] ||
		pixels[j][i+1].g != ColorMap[CM_GREEN][1] ||
		pixels[j][i+1].b != ColorMap[CM_BLUE][1])
		    return TRUE;
	}
    }
    return FALSE;
}

int
fclose_xpm_file()
{
    int i;

    XpmFreeXpmImage(&image);
    XpmFreeXpmInfo(&info);

    return FALSE;
}

#ifndef AMIGA
static char *std_args[] = { "tilemap",	/* dummy argv[0] */
			"monsters.xpm", "monsters.txt",
			"objects.xpm",  "objects.txt",
			"other.xpm",    "other.txt" };

int
main(argc, argv)
int argc;
char *argv[];
{
    pixel pixels[TILE_Y][TILE_X];

    if (argc == 1) {
	argc = SIZE(std_args);
	argv = std_args;
    } else if (argc != 3) {
	Fprintf(stderr, "usage: xpm2txt xpmfile txtfile\n");
	exit(EXIT_FAILURE);
    }

    while (argc > 1) {
	if (!fopen_xpm_file(argv[1], RDBMODE))
	    exit(EXIT_FAILURE);

	init_colormap();

	if (!fopen_text_file(argv[2], WRTMODE)) {
	    (void) fclose_xpm_file();
	    exit(EXIT_FAILURE);
	}

	while (read_xpm_tile(pixels))
	    (void) write_text_tile(pixels);

	(void) fclose_xpm_file();
	(void) fclose_text_file();

	argc -= 2;
	argv += 2;
    }
    exit(EXIT_SUCCESS);
    /*NOTREACHED*/
    return 0;
}
#endif
